@@ -6,12 +6,22 @@
 # Author: Philippe LABORIE, IBM Analytics, France Lab, Gentilly
 
 """
-This module contains the objects representing CPO function expressions
-(Stepwise and Piecewise Linear functions).
+This module contains the objects representing CP Optimizer function expressions,
+Stepwise and Piecewise Linear functions:
+
+In particular, it defines the following classes:
+
+ * :class:`CpoFunction`: the root class of all function expressions,
+ * :class:`CpoSegmentedFunction`: for functions represented as a list of segments,
+ * :class:`CpoStepFunction`: for functions represented as a list of steps.
+
+
+Detailed description
+--------------------
 """
 
 from docplex.cp.catalog import Type_SegmentedFunction, Type_StepFunction
-from docplex.cp.expression import CpoExpr, INT_MIN, INT_MAX
+from docplex.cp.expression import CpoExpr, INTERVAL_MIN, INTERVAL_MAX
 from docplex.cp.utils import *
 import bisect
 
@@ -75,6 +85,10 @@ class CpoFunction(CpoExpr):
             self._s = s
         assert len(self._s) == l, "Lists x and s must have the same length"
 
+    def _get_expr_string(self):
+        """ Get the string representing this expression, name excluded. """
+        return "Function(...)"
+
     @property
     def v0(self):
         return self._v0
@@ -97,11 +111,11 @@ class CpoFunction(CpoExpr):
 
     @property
     def is_step_function(self):
-        return self.get_type() == Type_StepFunction
+        return self.type == Type_StepFunction
 
     @property
     def is_segmented_function(self):
-        return self.get_type() == Type_SegmentedFunction
+        return self.type == Type_SegmentedFunction
 
     def get_value(self, t):
         """ Gets the value of the function.
@@ -135,7 +149,7 @@ class CpoFunction(CpoExpr):
         vcopy = self._v[:]
         scopy = self._s[:]
         if type is None:
-            tp = self.get_type()
+            tp = self.type
         else:
             tp = type
         return CpoFunction(tp, self._s0, self._v0, xcopy, vcopy, scopy)
@@ -163,7 +177,7 @@ class CpoFunction(CpoExpr):
                 self._x.append(x1)
                 self._v.append(v1)
                 self._s.append(s)
-                if x2 < INT_MAX:
+                if x2 < INTERVAL_MAX:
                     self._x.append(x2)
                     self._v.append(self._v0)
                     self._s.append(0)
@@ -267,7 +281,7 @@ class CpoFunction(CpoExpr):
             self._x.append(x1)
             self._v.append(self._v0 + v1)
             self._s.append(s)
-            if x2 < INT_MAX:
+            if x2 < INTERVAL_MAX:
                 self._x.append(x2)
                 self._v.append(self._v0)
                 self._s.append(0)
@@ -336,7 +350,6 @@ class CpoFunction(CpoExpr):
             self._v0 += ov0
             for i in range(len(self._x)):
                 self._v[i] += ov0
-            return
         else:
             ox0 = other._x[0]
             if ox0 == self._x[0]:
@@ -357,8 +370,8 @@ class CpoFunction(CpoExpr):
             for i in range(ol - 1):
                 # This could be optimized to avoid calling add_slope
                 self.add_slope(other._x[i], other._x[i + 1], other._v[i], other._s[i])
-            self.add_slope(other._x[ol - 1], INT_MAX, other._v[ol - 1], other._s[ol - 1])
-            return self
+            self.add_slope(other._x[ol - 1], INTERVAL_MAX, other._v[ol - 1], other._s[ol - 1])
+        return self
 
     def _add_number(self, k):
         self._v0 += k
@@ -378,7 +391,7 @@ class CpoFunction(CpoExpr):
         if isinstance(other, CpoFunction):
             assert (self.is_segmented_function or not other.is_segmented_function)
             return self._add(other)
-        elif isinstance(other, (bool, int, float)):
+        elif is_number(other) or is_bool(other):
             return self._add_number(other)
 
     def __isub__(self, other):
@@ -386,7 +399,7 @@ class CpoFunction(CpoExpr):
             assert (self.is_segmented_function or not other.is_segmented_function)
             tmp = other.copy()._mul(-1)
             return self._add(tmp)
-        elif isinstance(other, (bool, int, float)):
+        elif is_number(other) or is_bool(other):
             return self._add_number(-other)
 
     def __add__(self, other):
@@ -395,7 +408,7 @@ class CpoFunction(CpoExpr):
                 return other.copy(Type_SegmentedFunction)._add(self)
             else:
                 return other.copy()._add(self)
-        elif isinstance(other, (bool, int, float)):
+        elif is_number(other) or is_bool(other):
             return self.copy()._add_number(other)
 
     def __radd__(self, value):
@@ -407,7 +420,7 @@ class CpoFunction(CpoExpr):
                 return other.copy(Type_SegmentedFunction)._mul(-1)._add(self)
             else:
                 return other.copy()._mul(-1)._add(self)
-        elif isinstance(other, (bool, int, float)):
+        elif is_number(other) or is_bool(other):
             return self.copy()._add_number(-other)
 
     def __neg__(self):
@@ -453,9 +466,20 @@ class CpoFunction(CpoExpr):
 
 
 class CpoSegmentedFunction(CpoFunction):
-    def __init__(self, segment0=None, segments=None, name=None):
-        """ Segmented function.
+    """ Class representing a segmented function.
 
+    In CP Optimizer, piecewise linear functions are typically used in modeling a known function of time,
+    for instance the cost that is incurred for completing an activity after a known date.
+
+    A segmented function is a piecewise linear function defined on an interval [xmin, xmax)
+    which is partitioned into segments such that over each segment, the function is linear.
+
+    When two consecutive segments of the function are colinear, these segments are merged so that the function
+    is always represented with the minimal number of segments.
+    """
+
+    def __init__(self, segment0=None, segments=None, name=None):
+        """
         Args:
             segment0 (tuple): Initial segment of the function (slope, vright).
             segments (list): Segments of the function represented as a list of
@@ -474,32 +498,53 @@ class CpoSegmentedFunction(CpoFunction):
                 x.append(seg[0])
                 v.append(seg[1])
                 s.append(seg[2])
-        super(CpoSegmentedFunction, self).__init__(typ=Type_SegmentedFunction, s0=s0, v0=v0, x=x, v=v, s=s, name=name)
+        super(CpoSegmentedFunction, self).__init__(Type_SegmentedFunction, s0=s0, v0=v0, x=x, v=v, s=s, name=name)
 
     def get_segment_list(self):
         """ Returns the list of segments of the function.
         """
-        segments = []
-        segments.append((self.s0, self.v0))
+        segments = [(self.s0, self.v0)]
         for i in range(len(self.x)):
             segments.append((self.x[i], self.v[i], self.s[i]))
         return segments
 
+    def __str__(self):
+        """ Build a string representing this function. """
+        res = "SegmentedFunction(" + str(self.get_segment_list()) + ")"
+        if self.name:
+            res = self.name + "=" + res
+        return res
+
 
 class CpoStepFunction(CpoFunction):
+    """ Class representing a step function.
+
+    In CP Optimizer, stepwise functions are typically used to model the efficiency of a resource over time.
+
+    A stepwise function is a special case of piecewise linear function where all slopes are equal to 0 and the
+    domain and image of the function are integer.
+
+    When two consecutive steps of the function have the same value, these steps are merged so that the function
+    is always represented with the minimal number of steps.
+
+    The function steps are expressed as a list of couples (x, val) specifying that the value of the function is *val*
+    after *x*, up to the next step.
+    By default, the value of the function is zero up to the first step.
+    To change this default value, the first step should be (INTERVAL_MIN, value).
+    """
     def __init__(self, steps=None, name=None):
         """ Step function.
 
         Args:
-            steps:  Function steps
-            name: Function name
+            steps: (Optional) Function steps, expressed as a list of couples (x, val).
+            name:  (Optional) Function name
         """
         if steps is None or len(steps) == 0:
-            super(CpoStepFunction, self).__init__(typ=Type_StepFunction, name=name)
+            super(CpoStepFunction, self).__init__(Type_StepFunction, name=name)
         else:
             x = []
             v = []
-            if steps[0][0] == INT_MIN:
+            if steps[0][0] == INTERVAL_MIN:
                 v0 = steps[0][1]
             else:
                 v0 = 0
@@ -515,7 +560,16 @@ class CpoStepFunction(CpoFunction):
         """
         steps = []
         if self.v0 != 0:
-            steps.append((INT_MIN, self.v0))
+            steps.append((INTERVAL_MIN, self.v0))
         for i in range(len(self.x)):
             steps.append((self.x[i], self.v[i]))
         return steps
+
+    def __str__(self):
+        """ Build a string representing this function. """
+        res = "StepFunction(" + str(self.get_step_list()) + ")"
+        if self.name:
+            res = self.name + "=" + res
+        return res
+
+
