@@ -15,15 +15,13 @@ import sys
 
 from sys import version_info
 
-ERROR_STRING = "docplex is not compatible with this version of Python: only 64 bits on Windows, Linux and Darwin, with Python 2.7.9+ and 3.4.x are supported."
+ERROR_STRING = "docplex is not compatible with this version of Python: only 64 bits on Windows, Linux, Darwin and AIX, with Python 2.7.9+, 3.4+ are supported."
 
 platform_system = platform.system()
-if platform_system in ('Darwin', 'Linux', 'Windows', 'Microsoft'):
+if platform_system in ('Darwin', 'Linux', 'Windows', 'Microsoft', 'AIX'):
     if version_info[0] == 3:
         if version_info < (3, 4, 0):
             raise Exception(ERROR_STRING)
-        elif version_info >= (3, 5, 0):
-            warnings.warn("docplex is not officially supported on Python 3.5+. Use it at your own risk.", RuntimeWarning)
     elif version_info[0] == 2:
         if version_info[1] != 7:
             raise Exception(ERROR_STRING)
