@@ -7,7 +7,7 @@
 # gendoc: ignore
 
 # This file contains all compatibility stuff between py2/py3
-
+import sys
 
 # py2/py3 compatibility
 try:
@@ -36,6 +36,7 @@ except ImportError:                 # pragma: no cover
 
 
 try:
+    # noinspection PyCompatibility
     xrange(2)
     fast_range = xrange
 except NameError:       # pragma: no cover
@@ -43,5 +44,17 @@ except NameError:       # pragma: no cover
 
 try:
     from itertools import izip
-except ImportError:  # pragma : no cover
-    izip = zip       # pragma : no cover
+except ImportError:  # pragma: no cover
+    izip = zip       # pragma: no cover
+
+# we want unicode in py2, str otherwise
+# unitext() returns a unicode representation of its parameter
+if sys.version_info[0] == 3:  # pragma: no cover
+    unitext = str
+else:
+    unitext = unicode
+
+try:  # pragma: no cover
+    from itertools import zip_longest as izip_longest
+except ImportError:  # pragma: no cover
+    from itertools import izip_longest
\ No newline at end of file
