@@ -8,24 +8,66 @@
 # gendoc: ignore
 
 import json
+import os
+
+try:
+    # python 2
+    from urlparse import urlparse
+except ImportError:
+    # python 3
+    from urllib.parse import urlparse
+
 
 from datetime import datetime
 from six import iteritems, string_types
 from requests.exceptions import ConnectionError
 
+import warnings
+
+from docplex.util.status import JobSolveStatus
+
 from docloud.job import JobClient, DOcloudInterruptedException, DOcloudNotFoundError
-from docloud.status import JobSolveStatus, JobExecutionStatus
+from docloud.status import JobExecutionStatus
 
-from docplex.mp.progress import ProgressData
 from docplex.mp.utils import resolve_pattern, get_logger
 from docplex.mp.utils import CyclicLoop
 
+from docplex.mp.context import has_credentials
+
 
 def key_as_string(key):
     """For keys, we don't want the key to appear in INFO log outputs.
     Instead, we display the first 4 chars and the last 4 chars.
     """
-    return (key[:4]+"*******"+ key[-4:]) if isinstance(key, string_types) else str(key)
+    return (key[:4] + "*******" + key[-4:]) if isinstance(key, string_types) else str(key)
+
+
+def get_cplex_version(context):
+    '''Submits a minimal job on DOCplexcloud and get the CPLEX version from
+    logs
+    Args:
+        context: The context to use to connect
+    Returns:
+        A tuple (major, minor, micro), for instance (12, 7, 1)
+    '''
+    import re
+    from docplex.mp.compat23 import StringIO
+    from docplex.mp.model import Model
+    logs = StringIO()
+    model = Model()
+    model.solve(url=context.solver.docloud.url,
+                key=context.solver.docloud.key,
+                log_output=logs)
+    v = re.search('CPLEX version ([0-9][0-9])([0-9][0-9])([0-9][0-9])([0-9][0-9])', logs.getvalue())
+    try:
+        major = int(v.group(1))
+        minor = int(v.group(2))
+        micro = int(v.group(3))
+        return (major, minor, micro)
+    except IndexError:
+        raise RuntimeError('Could not find CPLEX version in resulting logs')
+    except:
+        raise ValueError('Could not determine version from %s' % v.group(0))
 
 
 class DOcloudConnectorException(Exception):
@@ -52,17 +94,18 @@ class DOcloudEvent(object):
 
 
 class DOcloudConnector(object):
-    # json keys
-    JSON_LINEAR_CTS_KEY = 'linearConstraints'
 
     def __init__(self, docloud_context, log_output=None):
         """ Starts a connector which URL and authorization are stored in the
         specified context, along with other connection parameters
 
+        The `docloud_context` refers to the context.solver.docloud node of a
+        context.
+
         Args:
             log_output: The log output stream
         """
-        if docloud_context is None or not docloud_context.has_credentials():
+        if docloud_context is None or not has_credentials(docloud_context):
             raise DOcloudConnectorException("Please provide DOcplexcloud credentials")
 
         # store this for future usage
@@ -79,16 +122,12 @@ class DOcloudConnector(object):
         self.logger.info("   waittime = " + str(docloud_context.waittime))
         self.logger.info("   timeout = " + str(docloud_context.timeout))
 
-        self.json = ""
         self.jobInfo = None
         self.run_deterministic = docloud_context.run_deterministic
         self.log_output = log_output
-        self._hasSolution = False  # submit_model_data change this
-        self.__vars = None
 
-    def _check_nonempty_json(self):
-        if not self.json:
-            raise DOcloudConnectorException("* empty JSON result!")
+        self.timed_out = False
+        self.results = {}
 
     def log(self, msg, *args):
         if self.docloud_context.verbose:
@@ -106,8 +145,8 @@ class DOcloudConnector(object):
     def submit_model_data(self,
                           attachments=None,
                           gzip=False,
-                        info_callback=None,
-                        info_to_monitor=None):
+                          info_callback=None,
+                          info_to_monitor=None):
         """Submits a job to the cloud service.
 
         Args:
@@ -122,21 +161,39 @@ class DOcloudConnector(object):
             info_to_monitor: A set of information to monitor with info_callback.
                 Currently, can be ``jobid`` and ``progress``.
         """
-        self._hasSolution = False
         self.__vars = None
+        self.timed_out = False
+        self.results.clear()
 
         if not info_to_monitor:
             info_to_monitor = {}
 
-        client = JobClient(self.docloud_context.url,
-                           self.docloud_context.key)
+        # check that url is valid
+        parts = urlparse(self.docloud_context.url)
+        if not parts.scheme:
+            raise DOcloudConnectorException("Malformed URL: '%s': No schema supplied." % self.docloud_context.url)
+
+        proxies = self.docloud_context.proxies
+        try:
+            client = JobClient(self.docloud_context.url,
+                               self.docloud_context.key,
+                               proxies=proxies)
+        except TypeError:
+            # docloud client <= 1.0.172 do not have the proxes
+            warnings.warn("Using a docloud client that do not support warnings in init()",
+                          UserWarning)
+            client = JobClient(self.docloud_context.url,
+                               self.docloud_context.key)
+        self.log("client created")
+        if proxies:
+            self.log("proxies = %s" % proxies)
 
         # prepare client
         if self.docloud_context.log_requests:
             client.rest_callback = \
                 lambda m, u, *a, **kw: self._rest_callback(m, u, *a, **kw)
         client.verify = self.docloud_context.verify
-        client.timeout = self.docloud_context.timeout if self.docloud_context.timeout is not None else None
+        client.timeout = self.docloud_context.get('timeout', None)
 
         try:
             try:
@@ -144,7 +201,8 @@ class DOcloudConnector(object):
                 att_names = [a['name'] for a in attachments]
 
                 # create job
-                jobid = client.create_job(attachments=att_names)
+                jobid = client.create_job(attachments=att_names,
+                                          parameters=self.docloud_context.job_parameters)
                 self.log("job creation submitted, id is: {0!s}".format(jobid))
                 if info_callback and 'jobid' in info_to_monitor:
                     info_callback({'jobid': jobid})
@@ -154,11 +212,33 @@ class DOcloudConnector(object):
             try:
                 # now upload data
                 for a in attachments:
+                    pos = 0
+                    if 'data' in a:
+                        att_data = {'data': a['data']}
+                    elif 'file' in a:
+                        att_data = {'file': a['file']}
+                        pos = a['file'].tell()
+                    elif 'filename' in a:
+                        att_data = {'filename': a['filename']}
+
                     client.upload_job_attachment(jobid,
                                                  attid=a['name'],
-                                                 data=a['data'])
+                                                 **att_data)
                     self.log("Attachment: %s has been uploaded" % a['name'])
-
+                    if self.docloud_context.debug_dump_dir:
+                        target_dir = self.docloud_context.debug_dump_dir
+                        if not os.path.exists(target_dir):
+                            os.makedirs(target_dir)
+                        self.log("Dumping input attachment %s to dir %s" % (a['name'], target_dir))
+                        with open(os.path.join(target_dir, a['name']), "wb") as f:
+                            if 'data' in 'a':
+                                if isinstance(a['data'], bytes):
+                                    f.write(a['data'])
+                                else:
+                                    f.write(a['data'].encode('utf-8'))
+                            else:
+                                a['file'].seek(pos)
+                                f.write(a['file'])
                 # execute job
                 client.execute_job(jobid)
                 self.log("DOcplexcloud execute submitted has been started")
@@ -174,31 +254,48 @@ class DOcloudConnector(object):
                 self.log("docloud execution has finished")
                 # get job status. Do this before any time out handling
                 self.jobInfo = client.get_job(jobid)
+
+                if self.docloud_context.fire_last_progress and info_callback:
+                    progress_data = self.map_job_info_to_progress_data(self.jobInfo)
+                    info_callback({'progress': progress_data})
+
                 if timedout:
-                    self._hasSolution = False
+                    self.timed_out = True
                     self.log("Solve timed out after {waittime} sec".format(waittime=self.docloud_context.waittime))
-                    return None
-                # get solution
+                    return
+                # get solution => download all attachments
                 try:
-                    solution_as_string = self._as_string(client.download_job_attachment(jobid, attid="solution.json"))
-                    myjson = json.loads(solution_as_string, parse_constant='utf-8')['CPLEXSolution']
-                    self._hasSolution = bool(myjson)
+                    for a in client.get_job_attachments(jobid):
+                        if a['type'] == 'OUTPUT_ATTACHMENT':
+                            name = a['name']
+                            self.log("Downloading attachment '%s'" % name)
+                            attachment_as_string = self._as_string(client.download_job_attachment(jobid,
+                                                                                                  attid=name))
+                            self.results[name] = attachment_as_string
+                            if self.docloud_context.debug_dump_dir:
+                                target_dir = self.docloud_context.debug_dump_dir
+                                if not os.path.exists(target_dir):
+                                    os.makedirs(target_dir)
+                                self.log("Dumping attachment %s to dir %s" % (name, target_dir))
+                                with open(os.path.join(target_dir, name), "wb") as f:
+                                    f.write(attachment_as_string.encode('utf-8'))
                 except DOcloudNotFoundError:
-                    myjson = None
-                    self._hasSolution = False
                     self.log("no solution in attachment")
                 self.log("docloud results have been received")
-                self.json = myjson
-                return myjson
-
+                # on_solve_finished_cb
+                if self.docloud_context.on_solve_finished_cb:
+                    self.docloud_context.on_solve_finished_cb(jobid=jobid,
+                                                              client=client,
+                                                              connector=self)
+                return
             finally:
-                deleted = client.delete_job(jobid)
-                self.log("delete status for job: {0!s} = {1!s}".format(jobid, deleted))
+                if self.docloud_context.delete_job:
+                    deleted = client.delete_job(jobid)
+                    self.log("delete status for job: {0!s} = {1!s}".format(jobid, deleted))
 
         finally:
             client.close()
 
-
     def wait_for_completion(self, client, jobid,
                             info_callback=None, info_to_monitor=None):
         def status_poll(loop):
@@ -240,7 +337,7 @@ class DOcloudConnector(object):
                     level = r['level'][:4]
                     date = r['date']
                     message = r['message'].rstrip()
-                    d = datetime.utcfromtimestamp(int(float(date)*0.001))
+                    d = datetime.utcfromtimestamp(int(float(date) * 0.001))
                     m = "[{date}Z, {level}] {message}\n".format(date=d.isoformat(),
                                                                 level=level,
                                                                 message=message
@@ -264,8 +361,15 @@ class DOcloudConnector(object):
                 info_to_monitor: what info does the callback want to monitor
             """
             if 'progress' in info_to_monitor and info_callback:
+                logger = self.docloud_context.verbose_progress_logger
+                if logger:
+                    logger.info("polling progress")
                 info = loop.client.get_job(loop.jobid)
+                if logger:
+                    logger.info("job info: %s" % json.dumps(info, indent=3))
                 if 'details' in info:  # there are some info available
+                    if logger:
+                        logger.info("generating progress event using_threads = %s" % using_threads)
                     progress_data = self.map_job_info_to_progress_data(info)
                     if using_threads:
                         loop.event_queue.put(DOcloudEvent("progress", progress_data))
@@ -289,11 +393,11 @@ class DOcloudConnector(object):
 
         # interval to check job status
         status_poll_interval = client.nice
-        log_poll_interval = self.docloud_context.log_poll_interval
-        if not log_poll_interval:
+        log_poll_interval = self.docloud_context.get('log_poll_interval')
+        if log_poll_interval is None:
             log_poll_interval = client.nice * 3
-        progress_poll_interval = self.docloud_context.progress_poll_interval
-        if not progress_poll_interval:
+        progress_poll_interval = self.docloud_context.get('progress_poll_interval')
+        if progress_poll_interval is None:
             progress_poll_interval = client.nice * 3
 
         # The cyclic loop
@@ -349,73 +453,10 @@ class DOcloudConnector(object):
 
         return loop.status
 
-    def has_solution(self):
-        return self._hasSolution
-
     def get_cplex_details(self):
         if self.jobInfo:
             return self.jobInfo.get("details")
 
-    def is_mip(self):
-        return self.json and self.json['header']['solutionMethodString'] == 'mip'
-
-    def variable_values(self):
-        return self.get_variable_attr_map('value')
-
-    def variable_reduced_costs(self):
-        return self.get_variable_attr_map('reducedCost')
-
-    def get_variable_attr_map(self, json_attr_name):
-        assert json_attr_name
-        if not self.json:
-            return {}
-        else:
-            all_vars = self._getvars()
-            attr_map = {int(v['index']): float(v[json_attr_name]) for v in all_vars}
-            return attr_map
-
-    def cplex_index_name_map(self):
-        json_res = self.json
-        if json_res:
-            return {int(v['index']): v['name'] for v in self._getvars()}
-        else:
-            return {}
-
-    def _getvars(self):
-        if self.__vars is None:
-            self.__vars = self.json.get('variables', [])
-        return self.__vars
-
-    def variable_results(self):
-        if not self.json:
-            return {}, {}
-        else:
-            all_vars = self._getvars()
-            value_map = {int(v['index']): float(v['value']) for v in all_vars}
-            rc_map = {} if self.is_mip() else {int(v['index']): float(v['reducedCost']) for v in all_vars}
-            return value_map, rc_map
-
-    def constraint_results(self):
-        if not self.json or self.is_mip():
-            return {}, {}
-        else:
-            lincst_key = self.JSON_LINEAR_CTS_KEY
-            if lincst_key in self.json:
-                all_linear_cts = self.json[lincst_key]
-                dual_map = {int(v['index']): float(v['dual']) for v in all_linear_cts}
-                slack_map = {int(v['index']): float(v['slack']) for v in all_linear_cts}
-            else:
-                dual_map = slack_map = {}
-            return dual_map, slack_map
-
-    def get_status_id(self):
-        self._check_nonempty_json()
-        return int(self.json['header']['solutionStatusValue'])
-
-    def get_objective(self):
-        self._check_nonempty_json()
-        return float(self.json['header']['objectiveValue'])
-
     def get_solve_status(self):
         if 'solveStatus' in self.jobInfo:
             return JobSolveStatus[self.jobInfo['solveStatus']]
@@ -440,22 +481,24 @@ class DOcloudConnector(object):
         Returns:
             A ProgressData
         """
-
-        pg = ProgressData()
+        from docplex.mp.progress import ProgressData
+        has_incumbent = current_objective = best_bound = mip_gap = current_nb_iterations = current_nb_nodes = remaining_nb_nodes = 0
+        time = 0
         details = info.get('details')
         if details:
-            pg.current_objective = float(details.get('PROGRESS_CURRENT_OBJECTIVE',
-                                                     pg.current_objective))
-            pg.best_bound = float(details.get('PROGRESS_BEST_OBJECTIVE',
-                                              pg.best_bound))
+            current_objective = float(details.get('PROGRESS_CURRENT_OBJECTIVE', 0))
+            best_bound = float(details.get('PROGRESS_BEST_OBJECTIVE', 0))
+            mip_gap = None
             if 'PROGRESS_CURRENT_OBJECTIVE' in details and 'PROGRESS_BEST_OBJECTIVE' in details:
-                if pg.current_objective > 0:
-                    pg.mip_gap = abs(pg.current_objective - pg.best_bound) / pg.current_objective
-            pg.current_nb_nodes = int(details.get('cplex.nodes.processed',
-                                                  pg.current_nb_nodes))
-            pg.remaining_nb_nodes = int(details.get('cplex.nodes.left',
-                                                    pg.remaining_nb_nodes))
+                if current_objective > 0:
+                    mip_gap = abs(current_objective - best_bound) / current_objective
+            current_nb_nodes = int(details.get('cplex.nodes.processed', 0))
+            remaining_nb_nodes = int(details.get('cplex.nodes.left', 0))
             # assume that there's an incubent if ther's a gap
-            pg.has_incumbent = 'PROGRESS_GAP' in details
-        pg.time = ((info.get('updatedAt')) - int(info.get('startedAt'))) / 1000
+            has_incumbent = 'PROGRESS_GAP' in details
+        if 'startedAt' in info and 'updatedAt' in info:
+            time = ((info.get('updatedAt')) - int(info.get('startedAt'))) / 1000
+        pg = ProgressData(0, has_incumbent, current_objective, best_bound, mip_gap,
+                          0, current_nb_nodes, remaining_nb_nodes,
+                          time, 0)
         return pg
