@@ -8,10 +8,9 @@ from __future__ import print_function
 
 from six import iteritems
 
-from docplex.mp.utils import is_int, is_number
+from docplex.mp.utils import is_int, is_string
 from docplex.mp.compat23 import StringIO
-from docplex.mp.error_handler import docplex_fatal
-from docplex.mp.utils import RedirectedOutputContext
+from docplex.mp.error_handler import docplex_fatal, DOcplexException
 
 
 class ParameterGroup(object):
@@ -47,7 +46,7 @@ class ParameterGroup(object):
         """
         from copy import deepcopy
         return deepcopy(self)
-    
+
     def copy(self):
         return self.clone()
 
@@ -96,6 +95,9 @@ class ParameterGroup(object):
         """
         return len(self._subgroups)
 
+    def has_subgroups(self):
+        return len(self._subgroups) > 0
+
     def iter_subgroups(self):
         return iter(self._subgroups)
 
@@ -127,7 +129,7 @@ class ParameterGroup(object):
             group = group.parent_group
         return group
 
-    def qualified_name(self, sep=".", upto_root=True):
+    def qualified_name(self, sep=".", include_root=True):
         """ Computes a string with all the parents of the parameters.
 
         Example:
@@ -135,6 +137,7 @@ class ParameterGroup(object):
 
         Args:
             sep (string): The separator string. Default is ".".
+            include_root (flag): True if the root name is included.
 
         Returns:
             string: A string representation of the parameter hierarchy.
@@ -142,10 +145,10 @@ class ParameterGroup(object):
         self_parent = self._parent
         if not self_parent:
             return self.name
-        if not upto_root and self_parent.is_root():
+        if not include_root and self_parent.is_root():
             return self.name
         else:
-            return "".join([self._parent.qualified_name(sep=sep, upto_root=upto_root), sep, self.name])
+            return "".join([self._parent.qualified_name(sep=sep, include_root=include_root), sep, self.name])
 
     def prettyprint(self, indent=0):
         tab = indent * 4 * " "
@@ -164,9 +167,20 @@ class ParameterGroup(object):
             for k in extra_dict:
                 if k in self_dict:
                     # should not happen
-                    print("!! update_self_dict: name collision with: %s" % k)  # pragma : no cover
+                    print("!! update_self_dict: name collision with: %s" % k)  # pragma: no cover
         self_dict.update(extra_dict)
 
+    def _restore_dict_recursive(self):
+        # internal use for  deepcopy
+        param_dict = {p._short_name: p for p in self.iter_params()}
+        subgroup_dict = {sg._name: sg for sg in self.iter_subgroups()}
+        # update the __dict__ itself
+        self._update_self_dict(param_dict, do_check=False)
+        self._update_self_dict(subgroup_dict, do_check=False)
+        # recurse
+        for sg in self.iter_subgroups():
+            sg._restore_dict_recursive()
+
     @staticmethod
     def make(name, param_dict_fn, subgroup_fn, parent=None):
         # INTERNAL
@@ -232,8 +246,8 @@ class ParameterGroup(object):
 
         Args:
            predicate: A function that takes one parameter as asrgument.
-           The return value of this function will be interpeted as a boolean using
-           Python conversionb rules.
+           The return value of this function will be interpreted as a boolean using
+           Python conversion rules.
 
         Returns:
             A generator object.
@@ -271,7 +285,7 @@ class ParameterGroup(object):
         elif hasattr(self, attr_name):
             attr = getattr(self, attr_name)
             if isinstance(attr, Parameter):
-                # attribute is set inside param, not necesarily in engine...
+                # attribute is set inside param, not necessarily in engine...
                 attr.set(value)
             else:
                 docplex_fatal("No parameter with name {0} in {1}", attr_name, self.qualified_name())
@@ -289,6 +303,12 @@ class Parameter(object):
     This class is not meant to be instantiated but subclassed.
 
     """
+    __slots__ = ('_parent', '_short_name', '_cpx_name', '_id', '_description', '_default_value',
+                 '_current_value', '_synchronous')
+
+    # This global flag controls checking new values.
+    # If set to False, assigned values are not checked for min/max ranges
+    skip_range_check = False
 
     # noinspection PyProtectedMember
     def __init__(self, group, short_name, cpx_name, param_key, description, default_value):
@@ -301,15 +321,22 @@ class Parameter(object):
         self._default_value = default_value
         # current = default at start...
         self._current_value = default_value
+        self._synchronous = False
         # link to parent group
         group._add_param(self)
 
+    def ctor_name(self):
+        return self.__class__.__name__
+
+    def is_synchronous(self):
+        return self._synchronous
+
     @property
     def short_name(self):
         return self._short_name
 
     @property
-    def qualified_name(self, sep='.'):
+    def qualified_name(self):
         """ Returns a hierarchical name string for the parameter.
 
         The qualified name reflects the location of the parameter in the parameter hierarchy.
@@ -323,7 +350,14 @@ class Parameter(object):
         :rtype:
             string
         """
-        return "%s.%s" % (self._parent.qualified_name(sep=sep), self.short_name)
+        return self.get_qualified_name(sep='.', include_root=True)
+
+    def get_qualified_name(self, sep='.', include_root=True):
+        parent_qname = self._parent.qualified_name(sep=sep, include_root=include_root)
+        if parent_qname:
+            return "%s.%s" % (parent_qname, self.short_name)
+        else:
+            return self.short_name
 
     @property
     def cpx_name(self):
@@ -378,7 +412,7 @@ class Parameter(object):
         self._current_value = new_default  # pragma: no cover
 
     @property
-    def current_value(self):
+    def value(self):
         return self._current_value
 
     def accept_value(self, new_value):
@@ -401,11 +435,19 @@ class Parameter(object):
         if raw_value == self.default_value:
             return raw_value
         elif not self.accept_value(raw_value):
-            docplex_fatal("Value {0!s} of type {2} is invalid for parameter {1}",
-                          raw_value, self.qualified_name, type(raw_value))
+            docplex_fatal("Value {0!r} of type {2} is invalid for parameter '{1}'",
+                          raw_value, self.get_qualified_name(include_root=False), type(raw_value))
         else:
             return self.transform_value(raw_value)
 
+    def __call__(self, *args):
+        if not args:
+            return self._current_value
+        elif len(args) == 1:
+            self.set(args[0])
+        else:
+            docplex_fatal('Call parameter accepts either 0 or 1 argument')
+
     def set(self, new_value):
         """ Changes the value of the parameter to `new_value`.
 
@@ -422,9 +464,12 @@ class Parameter(object):
             An exception if the value is not valid.
         """
         accepted_value = self._check_value(new_value)
-        # do not use if accepted value here as 0 evalutes to False in Python!
         if accepted_value is not None:
             self._current_value = accepted_value
+            if self._synchronous:
+                #  print(" syncing {0!s} to {1}".format(self, accepted_value))
+                self.root_group().apply(self)
+
         return accepted_value
 
     def get(self):
@@ -454,12 +499,13 @@ class Parameter(object):
         """
         return self.get() == self.default_value
 
-    @staticmethod
-    def _is_in_range(arg, range_min, range_max):
-        if range_min is not None and arg < range_min:
-            return False
-        if range_max is not None and arg > range_max:
-            return False
+    @classmethod
+    def _is_in_range(cls, arg, range_min, range_max):
+        if not cls.skip_range_check:
+            if range_min is not None and arg < range_min:
+                return False
+            if range_max is not None and arg > range_max:
+                return False
         return True
 
     def to_string(self):
@@ -470,7 +516,7 @@ class Parameter(object):
         :rtype:
             string
         """
-        return "{0}:{1:s}(2!s)".format(self._short_name, self.type_name(), self._current_value)
+        return "{0}:{1:s}({2!s})".format(self._short_name, self.type_name(), self._current_value)
 
     def __str__(self):
         return self.to_string()
@@ -479,11 +525,12 @@ class Parameter(object):
         # INTERNAL
         return False  # pragma: no cover
 
+    @property
     def type_name(self):
         # INTERNAL
         raise NotImplementedError  # pragma: no cover
 
-    def _root_group(self):
+    def root_group(self):
         return self._parent.root_group()
 
     def _repr_classname(self):
@@ -498,6 +545,8 @@ _BOOLEAN_STATES = {'1': True, 'yes': True, 'true': True, 'on': True,
 
 
 class BoolParameter(Parameter):
+    __slots__ = ()
+
     def __init__(self, group, short_name, cpx_name, param_key, description, default_value):
         Parameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value)
 
@@ -515,11 +564,14 @@ class BoolParameter(Parameter):
     def accept_value(self, value):
         return value in {0, 1} or str(value).lower() in _BOOLEAN_STATES or value in {True, False}
 
+    @property
     def type_name(self):
         return "bool"
 
 
 class StrParameter(Parameter):
+    __slots__ = ()
+
     def __init__(self, group, short_name, cpx_name, param_key, description, default_value):
         Parameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value)
         assert isinstance(default_value, str)
@@ -527,14 +579,16 @@ class StrParameter(Parameter):
     def accept_value(self, new_value):
         return isinstance(new_value, str)
 
+    @property
     def type_name(self):
         return "string"
 
 
 class IntParameter(Parameter):
+    __slots__ = ('_min_value', '_max_value')
+
     def accept_value(self, new_value):
-        ivalue = int(new_value)
-        return is_int(ivalue) and self._is_in_range(ivalue, self._min_value, self._max_value)
+        return is_int(new_value) and self._is_in_range(new_value, self._min_value, self._max_value)
 
     def is_numeric(self):
         return True  # pragma: no cover
@@ -546,11 +600,13 @@ class IntParameter(Parameter):
         return self._max_value  # pragma: no cover
 
     def __init__(self, group, short_name, cpx_name, param_key, description, default_value, min_value=None,
-                 max_value=None):
+                 max_value=None, sync=False):
         Parameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value)
         self._min_value = min_value
         self._max_value = max_value
+        self._synchronous = sync
 
+    @property
     def type_name(self):
         return "int"
 
@@ -558,24 +614,25 @@ class IntParameter(Parameter):
         return "{0}({1},{2!s})".format(self._repr_classname(), self.qualified_name, self._current_value)
 
     def transform_value(self, new_value):
-        if is_int(new_value):
-            return new_value
-        else:
-            return int(new_value)
+        return int(new_value)
 
 
 class PositiveIntParameter(IntParameter):
+    __slots__ = ()
+
     def __init__(self, group, short_name, cpx_name, param_key, description, default_value, max_value=None):
         IntParameter.__init__(self, group, short_name, cpx_name, param_key, description, default_value, min_value=0,
                               max_value=max_value)
 
+    @property
     def type_name(self):
-        return "count"
+        return "positive int"
 
 
 class NumParameter(Parameter):
     """ A numeric parameter can take any floating-point value in the range of `[min,max]` values.
     """
+    __slots__ = ('_min_value', '_max_value')
 
     def __init__(self, group, short_name, cpx_name, param_key, description, default_value, min_value=None,
                  max_value=None):
@@ -597,13 +654,12 @@ class NumParameter(Parameter):
         return self._is_in_range(fvalue, self._min_value, self._max_value)
 
     def transform_value(self, new_value):
-        if is_number(new_value):
-            return new_value
-        else:
-            return float(new_value)
+        return float(new_value)
 
+    @property
     def type_name(self):
-        return "num"
+        return "number"
+
 
 # a dictionary of formats for each type.
 _param_prm_formats = {NumParameter: "%.14f",
@@ -621,9 +677,36 @@ class RootParameterGroup(ParameterGroup):
     def __init__(self, name, cplex_version):
         ParameterGroup.__init__(self, name)
         self._cplex_version = cplex_version
+        self._models = []
 
-    def qualified_name(self, sep=".", upto_root=True):
-        return self.name
+    def __deepcopy__(self, memodict={}):
+        """
+        Returns:
+           A deep copy of the parameter group.
+        """
+        import copy
+        saved_models = self._models
+        self._models = []
+        new_root = RootParameterGroup(self.name, self._cplex_version)
+        memodict[id(self)] = new_root
+        for p in self.iter_params():
+            new_root._add_param(copy.copy(p))
+        for sg in self.iter_subgroups():
+            new_root._add_subgroup(copy.deepcopy(sg, memodict))
+        # add parameter and subgroup names to the inner __dict__ of the root
+        new_root._restore_dict_recursive()
+        # ---
+        for m in saved_models:
+            new_root.connect_model(m)
+        return new_root
+
+    def connect_model(self, m):
+        self._models.append(m)
+
+    def apply(self, param):
+        # apply one parameter to connected models
+        for m in self._models:
+            m.apply_one_parameter(param)
 
     @property
     def cplex_version(self):
@@ -646,7 +729,7 @@ class RootParameterGroup(ParameterGroup):
         only non-default parameters are printed.
 
         Args:
-            output: The output stream, typically a filename.
+            out: The output stream, typically a filename.
 
             overload_params: A dictionary of overloaded values for
                 certain parameters. This dictionary is of the form {param: value}
@@ -673,7 +756,19 @@ class RootParameterGroup(ParameterGroup):
         out.write("# --- end of generated prm data ---\n")
 
     def print_information(self, indent_level=0, print_all=False):
-        self.print_info_to_stream(output=None, overload_params=None, indent_level=indent_level, print_defaults=print_all)
+        indent = ' ' * indent_level
+        param_generator = self.generate_params()
+        for param in param_generator:
+            if param.is_nondefault() or print_all:
+                param_value = param.get()
+                print("{0}{1} = {2!s}"
+                      .format(indent,
+                              param.qualified_name,
+                              _param_prm_formats[type(param)] % param_value))
+
+    def export_prm_to_path(self, path, overload_params=None):
+        with open(path, mode='w') as out:
+            self.export_prm(out, overload_params=overload_params)
 
     def print_info_to_stream(self, output, overload_params=None, print_defaults=False, indent_level=0):
         """ Writes parameters to an output stream.
@@ -693,20 +788,19 @@ class RootParameterGroup(ParameterGroup):
                 certain parameters. This dictionary is of the form {param: value}.
         """
         indent = " " * indent_level
-        with RedirectedOutputContext(output):
-            param_generator = self.generate_params()
-            for param in param_generator:
-                if overload_params and param in overload_params:
-                    param_value = overload_params[param]
-                elif print_defaults or param.is_nondefault():
-                    param_value = param.get()
-                else:
-                    param_value = None
-                if param_value is not None:
-                    print("{0}{1} = {2!s}"
-                          .format(indent,
-                                  param.qualified_name,
-                                  _param_prm_formats[type(param)] % param_value))
+        param_generator = self.generate_params()
+        for param in param_generator:
+            if overload_params and param in overload_params:
+                param_value = overload_params[param]
+            elif print_defaults or param.is_nondefault():
+                param_value = param.get()
+            else:
+                param_value = None
+            if param_value is not None:
+                output.write("{0}{1} = {2!s}\n"
+                             .format(indent,
+                                     param.qualified_name,
+                                     _param_prm_formats[type(param)] % param_value))
 
     def export_prm_to_string(self, overload_params=None):
         """  Exports non-default parameters in PRM format to a string.
@@ -769,8 +863,8 @@ class RootParameterGroup(ParameterGroup):
     def __repr__(self):
         return "docplex.mp.params.RootParameterGroup(%s)" % self.cplex_version
 
-        # def qualified_name(self, sep='.'):
-        # return "parameters"
+    def qualified_name(self, sep='.', include_root=True):
+        return 'parameters' if include_root else ''
 
     def as_dict(self):
         # INTERNAL: returns a dictionary of qualified name -> parameter
@@ -785,19 +879,42 @@ class RootParameterGroup(ParameterGroup):
             for nd in sg._generate_and_filter_params(predicate=None):
                 yield nd
 
-    def update(self, other_params):
-        if not isinstance(other_params, RootParameterGroup):
-            docplex_fatal("Parameter.update expects  RootParameterGroup, got: {0!s}", other_params)
-        elif self._cplex_version != other_params.cplex_version:
-            docplex_fatal("Parameter.update expects same CPLEX version, self: {0}, other: {1}",
-                          format(self.cplex_version, other_params.cplex_version))
+    def find_parameter(self, key):
+        if is_int(key):
+            pred = lambda p_: p_.cpx_id == key
+        elif is_string(key):
+            # eliminate initial '.'
+            pred = lambda p_: p_.get_qualified_name(include_root=False) == key
         else:
-            self_qdict = self.as_dict()
-            nb_updates = 0
-            for other_param in other_params:
-                self_param = self_qdict[other_param.qualified_name]
-                other_current = other_param.get()
-                if self_param.get() != other_current:
-                    self_param.set(other_current)
-                    nb_updates += 1
-            return nb_updates
+            docplex_fatal('Parameters.find() accepts either integer code or path-like name, got: {0!r}'.format(key))
+        for p in self:
+            if pred(p):
+                return p
+        else:
+            return None
+
+    def set_from_qualified_name(self, qname, pvalue):
+        qname_list = qname.split('.')
+        if not qname_list:
+            # empty qname
+            return
+        groups = qname_list[:-1]
+        pname = qname_list[-1]
+        group = self
+        for g in groups:
+            group = getattr(group, g)
+            if group is None:
+                raise ValueError("Bad parameter group name: {0}".format(g))
+        res = None
+        try:
+            # extract parameter from group
+            target_param = getattr(group, pname)
+            if target_param:
+                target_param.set(pvalue)
+                res = pvalue
+        except AttributeError:
+            raise ValueError("Cannot find paramater {0} in group {1}".format(pname, group))
+        except DOcplexException as dex:
+            raise
+
+        return res
\ No newline at end of file
