@@ -1,13 +1,16 @@
 # --------------------------------------------------------------------------
 # Source file provided under Apache License, Version 2.0, January 2004,
 # http://www.apache.org/licenses/
-# (c) Copyright IBM Corp. 2015, 2016
+# (c) Copyright IBM Corp. 2015, 2019
 # --------------------------------------------------------------------------
 
 # gendoc: ignore
 # This file is generated !
 # See script tools/gen_version.py
-docplex_version_major = 1
-docplex_version_minor = 0
-docplex_version_micro = 565
-docplex_version_string = '1.0.565'
+docplex_version_major = 2
+docplex_version_minor = 17
+docplex_version_micro = 199
+docplex_version_string = '2.17.199'
+
+latest_cplex_major = 12
+latest_cplex_minor = 10
